/*
 *	portscan match for iptables
 *	Copyright © Jan Engelhardt <jengelh@computergmbh.de>, 2006 - 2007
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License; either version
 *	2 or 3 as published by the Free Software Foundation.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <xtables.h>
#include <iptables.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_portscan.h>

static void portscan_help(void)
{
	printf(
		"portscan match v%s options:\n"
		"(Combining them will make them match by OR-logic)\n"
		"  --stealth    Match TCP Stealth packets\n"
		"  --synscan    Match TCP SYN scans\n"
		"  --cnscan     Match TCP Connect scans\n"
		"  --grscan     Match Banner Grabbing scans\n",
		IPTABLES_VERSION);
	return;
}

static void portscan_mtinit(struct xt_entry_match *match, unsigned int *nfcache)
{
	/* Cannot cache this */
	*nfcache |= NFC_UNKNOWN;
	return;
}

static int portscan_parse(int c, char **argv, int invert, unsigned int *flags,
    const void *entry, unsigned int *nfcache, struct xt_entry_match **match)
{
	struct xt_portscan_info *info = (void *)((*match)->data);

	switch(c) {
		case 'c':
			info->match_cn = 1;
			return 1;
		case 'g':
			info->match_gr = 1;
			return 1;
		case 's':
			info->match_syn = 1;
			return 1;
		case 'x':
			info->match_stealth = 1;
			return 1;
		default:
			return 0;
	}
}

static void portscan_check(unsigned int flags)
{
	return;
}

static void portscan_print(const void *ip, const struct xt_entry_match *match,
    int numeric)
{
	const struct xt_portscan_info *info = (const void *)(match->data);
	const char *s = "";

	printf("portscan ");
	if(info->match_stealth) {
		printf("STEALTH");
		s = ",";
	}
	if(info->match_syn) {
		printf("%sSYNSCAN", s);
		s = ",";
	}
	if(info->match_cn) {
		printf("%sCNSCAN", s);
		s = ",";
	}
	if(info->match_gr)
		printf("%sGRSCAN", s);
	printf(" ");
	return;
}

static void portscan_save(const void *ip, const struct xt_entry_match *match)
{
	const struct xt_portscan_info *info = (const void *)(match->data);
	if(info->match_stealth)	printf("--stealth ");
	if(info->match_syn)	printf("--synscan ");
	if(info->match_cn)	printf("--cnscan ");
	if(info->match_gr)	printf("--grscan ");
	return;
}

static const struct option portscan_opts[] = {
	{"stealth", 0, NULL, 'x'},
	{"synscan", 0, NULL, 's'},
	{"cnscan",  0, NULL, 'c'},
	{"grscan",  0, NULL, 'g'},
	{NULL},
};

static struct xtables_match portscan_reg = {
	.name          = "portscan",
	.version       = IPTABLES_VERSION,
	.family        = AF_INET,
	.size          = XT_ALIGN(sizeof(struct xt_portscan_info)),
	.userspacesize = XT_ALIGN(sizeof(struct xt_portscan_info)),
	.help          = portscan_help,
	.init          = portscan_mtinit,
	.parse         = portscan_parse,
	.final_check   = portscan_check,
	.print         = portscan_print,
	.save          = portscan_save,
	.extra_opts    = portscan_opts,
};

void _init(void)
{
	xtables_register_match(&portscan_reg);
	return;
}
